package sorting;

import java.text.NumberFormat;
import java.util.ArrayList;

/**
 * A class that contains several sorting routines, implemented as static
 * methods. Arrays are rearranged with smallest item first, using compareTo.
 * Show intermediate results with annotation.
 * 
 * @author Mark Allen Weiss (modified by Ron Mak)
 */
public final class CountSort 
{    
    /**
     * Make a copy of an array of Integer.
     * @param a the array to copy.
     * @return the copy.
     */
    private static Integer[] copyOf(Integer a[])
    {
        Integer copy[] = new Integer[a.length];
        for (int i = 0; i < a.length; i++) copy[i] = a[i];
        return copy;
    }
    
    /**
     * Check to make sure the array is sorted.
     * @param a the array to check.
     * @return true if sorted, else false.
     */
    private static boolean checkSorted(Integer a[])
    {
        for (int i = 1; i < a.length; i++) {
            if (a[i-1] > a[i]) return false;
        }
        
        return true;
    }
    
    /**
     * Check that an array was properly sorted.
     * Print the sorting statistics.
     * @param a the array that was sorted.
     * @param stats the stats to print.
     */
    private static void printStats(Integer a[], Stats stats)
    {
        if (!checkSorted(a)) {
            System.out.println("    *** SORT ERROR ***");
        }
        else {
            System.out.printf("%15s%15s%15s\n",
                NumberFormat.getIntegerInstance().format(stats.moves), 
                NumberFormat.getIntegerInstance().format(stats.compares), 
                NumberFormat.getIntegerInstance().format(stats.time));
        }
    }
    
    /**
     * Print an array.
     * @param a the array to print.
     * @param p the array of positions to mark.
     */
    private static void printArray(Integer[] a)
    {
        System.out.println();
        for (Integer value : a) {
            System.out.printf(" %d", value);
        }
        System.out.println();
    }
    
    /**
     * Perform a sort on an array or a linked list using a sorting algorithm.
     * @param title the title of the sorting algorithm.
     * @param array the array to sort.
     * @param aFlag true if sorting an array, false if sorting a linked list.
     * @param algorithm the sorting algorithm.
     */
    private static void doSort(String title, Integer array[], boolean aFlag,
                               SortingAlgorithm<Integer> algorithm)
    {
        System.out.printf("%25s", title);
        Integer a[] = copyOf(array);
        //printArray(a);
        
        // If sorting a linked list, create one.
        MyLinkedList<Integer> list = aFlag ? null : new MyLinkedList<>(a);
        
        // Do the timed sort.
        long now = System.currentTimeMillis();
        Stats stats = aFlag ? algorithm.sort(a) : algorithm.sort(list);
        stats.time = System.currentTimeMillis() - now;
        
        // Copy the sorted linked list back into the array.
        if (!aFlag) {
            MyNode<Integer> node = list.head();
            for (int i = 0; i < a.length; i++) {
                a[i] = node.data;
                node = node.next;
            }
        }
        
        //printArray(a);
        printStats(a, stats);
    }
    
    /**
     * Run the sorting test.
     * @param title the test title.
     * @param generator the generator of the array to sort.
     */
    public static void run(String title, ArrayGenerator generator)
    {
        System.out.println();
        for (int i = 0; i < title.length(); i++) System.out.print("=");
        System.out.printf("\n%s\n", title);
        for (int i = 0; i < title.length(); i++) System.out.print("=");
        System.out.println();
        
        for (int N = 100; N <= 100000; N *= 10) {
            System.out.printf("\nN = %s\n\n", 
                              NumberFormat.getIntegerInstance().format(N));
            System.out.printf("%25s%15s%15s%15s\n", 
                              "ALGORITHM", "MOVES", "COMPARES", "MILLISECONDS");
            
            // Generate an array to sort.
            Integer array[] = generator.generate(N);
            
            // Do each sorting algorithm.
            doSort("Insertion sort",         
                    array, true, new InsertionSort());
            doSort("Shellsort suboptimal",   
                    array, true, new ShellSortSubOptimal());
            doSort("Shellsort Knuth",        
                    array, true, new ShellSortKnuth());
//*
            doSort("Heap sort",              
                    array, true, new HeapSort());
            doSort("Merge sort array",       
                    array, true, new MergeSortArray());
            doSort("Merge sort linked list", 
                    array, false, new MergeSortLinkedList());
            doSort("Quicksort suboptimal",              
                    array, true, new QuickSort(new PivotFirst()));
            doSort("Quicksort optimal",              
                    array, true, new QuickSort(new PivotMedianOfThree()));
//*/
        }
    }

    /**
     * Main.
     * @param args the command-line arguments (ignored).
     */
    public static void main(String[] args) 
    {     
        // Run the sorting tests with different generated arrays.
        run("Unsorted Random Unique", new GenerateUnsortedUnique());
        run("Sorted Unique",          new GenerateSortedUnique());
        run("Reverse Sorted Unique",  new GenerateReverseSortedUnique());
        run("All Zeroes",             new GenerateAllZeroes());
    }
}
